/**
  *  \file game/act-structure.cc
  *  \brief Building structures on a planet
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/assert.h"
#include "game/action.h"
#include "game/error.h"
#include "util/nls.h"

/********************** GPlanetBuildStructuresAction *********************/

/*! \class GPlanetBuildStructuresAction

    This class allows you to build structures on a planet. The cost
    can be billed to any container. The class enforces that maximum
    limits are never exceeded (i.e. it will never go into an error
    state because of exceeded building limits, it will simply never
    let you build so much), but you can build more than you have money
    for (to compute the price). */

/** Constructor.
    \param the_planet planet we are building on
    \param container  will pay for the structures
    \pre we are playing the planet and the container */
GPlanetBuildStructuresAction::GPlanetBuildStructuresAction(GPlanet& the_planet,
                                                           GCargoContainer& container)
    : GCargoCostTransaction(container),
      the_planet(the_planet),
      arb(the_planet.getCargoHandler(), GPlanet::pf_Structures)
{
    if (!arb)
        throw GError(GError::eFacility, _("Structure building already in progress on this planet"));
    for (int i = 0; i < MAX; ++i) {
        TPlanetaryStructure ps = TPlanetaryStructure(i);
        data[ps].order = data[ps].min = the_planet.getStructures(ps);
        data[ps].max = the_planet.getMaxStructures(ps);
    }
}

/** Set undo information. Enables this transaction to undo former
    GPlanetBuildStructuresAction's. */
void
GPlanetBuildStructuresAction::setUndoInformation(const GPlanet& dis_planet)
{
    for (int i = 0; i < MAX; ++i) {
        int dis_count = dis_planet.getStructures(TPlanetaryStructure(i));
        if (dis_count < data[i].min)
            data[i].min = dis_count;
    }
}

/** Destructor. */
GPlanetBuildStructuresAction::~GPlanetBuildStructuresAction()
{ }

/** Recompute costs. */
void
GPlanetBuildStructuresAction::computeCost()
{
    int nmines = data[str_Mines].order - the_planet.getStructures(str_Mines),
        nfact  = data[str_Factories].order - the_planet.getStructures(str_Factories),
        ndef   = data[str_Defense].order - the_planet.getStructures(str_Defense),
        nsbd   = data[str_BaseDefense].order - the_planet.getStructures(str_BaseDefense);

    setCost(el_Money,    4 * nmines + 3 * nfact + 10 * (ndef + nsbd));
    setCost(el_Supplies, nmines + nfact + ndef);
    setCost(el_Duranium, nsbd);
}

/** Add structures. This routine will not add or remove more than allowed by the
    rules, but it can add more structures than there is cash for. In that case,
    the GPlanetBuildStructuresAction will go into error state.
    \param ps      structure type
    \param count   amount to add (negative: remove)
    \param partial allow partial add

    \return number of structures added. With partial=false, return
    value is guaranteed to be either 0 or count. */
int
GPlanetBuildStructuresAction::addStructures(TPlanetaryStructure ps, int count, bool partial)
{
    if (count > 0) {
        /* add things */
        if (data[ps].order + count > data[ps].max) {
            if (!partial)
                return 0;
            count = data[ps].max - data[ps].order;
        }
    } else {
        /* remove things */
        if (data[ps].order + count < data[ps].min) {
            if (!partial)
                return 0;
            count = data[ps].min - data[ps].order;
        }
    }

    data[ps].order += count;
    computeCost();
    return count;
}

/** Add structures limited by cash. Like addStructures(), but will not add more than
    resources allow. If already over-allocated, this will not automatically turn an
    'add' order into a removal.
    \param ps      structure type
    \param count   amount to add (negative to remove) */
int
GPlanetBuildStructuresAction::addStructuresLimitCash(TPlanetaryStructure ps, int count)
{
    /* check how much we can add according to the rules */
    int rule_count = addStructures(ps, count, true);

    /* okay, we can add that much, but we don't have the cash. */
    while (rule_count > 0 && !isValidCost()) {
        addStructures(ps, -1, true);
        rule_count--;
    }
    return rule_count;
}

/** Commit transaction. */
void
GPlanetBuildStructuresAction::commit()
{
    ASSERT(!getError());
    for (int i = 0; i < MAX; ++i) {
        ASSERT(data[i].order <= data[i].max);
        ASSERT(data[i].order >= data[i].min);
        the_planet.setStructures(TPlanetaryStructure(i), mp16_t(data[i].order));
    }
    GCargoCostTransaction::commit();
}
