/**
  *  \file game/spec.cc
  *  \brief Specification files.
  *
  *  This file contains the classes which correspond to static game
  *  information (hullspec etc.), and functions to load/access the
  *  static specification files.
  *
  *  \todo Management of short names
  *  \todo Free-form field for starship hulls?
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/spec.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "game/gamedirs.h"
#include "io/iostr.h"
#include "util/console.h"
#include "util/confpars.h"
#include "util/misc.h"
#include "game/config.h"
#include "cpluslib/format.h"
#include "game/host.h"

/** Beam specifications. Indexed 0 to NUM_BEAMS-1 */
std::vector<GBeam> beam_spec;

/** Torpedo specifications. Indexed 0 to NUM_TORPS-1 */
std::vector<GTorpLauncher> torp_spec;

/** Engine specifications. Indexed 0 to NUM_ENGINES-1 */
std::vector<GEngine> eng_spec;

/** Hull specification. Indexed 0 to NNN */
std::vector<GHull*> hull_spec;

/** Truehull. For each player the hulls he can build. */
TTruehull truehull_spec;

/** \class GComponent
    \brief Component of a game item

    This class represents an abstract component of something
    (weapon, structure, ...) */
GComponent::~GComponent()
{ }

/********************************* GBeam *********************************/

/** \class GBeam
    \brief Beam Weapon

    This class represents a beam weapon. */
GBeam::GBeam(int aid, const TBeam& adata)
    : GWeapon(aid), data(adata)
{ }

/******************************** GTorpedo *******************************/

/** \class GTorpedo
    \brief Torpedo

    This class represents a torpedo. It differs from GTorpLauncher
    only in the cost it returns. */
GTorpedo::GTorpedo(int id, const TTorpedo& data)
    : GWeapon(id), data(data)
{ }

TCost
GTorpedo::getCost() const
{
    TCost c;
    c.mc = data.torp_cost;
    c.t = c.d = c.m = 1;
    return c;
}

/***************************** GTorpLauncher *****************************/

/** \class GTorpLauncher
    \brief Torpedo Launcher

    This class represents a torpedo launcher. */
GTorpLauncher::GTorpLauncher(int id, const TTorpedo& data)
    : GWeapon(id), torp(id, data)
{ }

/******************************** GFighter *******************************/

/** \class GFighter
    \brief Fighter

    This class represents a fighter. Its main purpose is to provide a
    container for the fighter specs. Unlike the other specs, which
    represent different types which are constant for all players,
    there is only one type of fighters, but that one can differ for
    all players. */
string_t
GFighter::getName() const
{
    return _("Fighter");
}

int
GFighter::getKill() const
{
    if (host.isPHost())
        return CONFIG_ARRAY(FighterBeamKill, getId());
    else
        return 2;
}

int
GFighter::getDamage() const
{
    if (host.isPHost())
        return CONFIG_ARRAY(FighterBeamExplosive, getId());
    else
        return 2;
}

TCost
GFighter::getCost() const
{
    TCost c;
    c.mc = 100;
    c.t  = 3;
    c.d  = 0;
    c.m  = 5;
    return c;
}

/******************************** GEngine ********************************/

/** \class GEngine
    \brief Engine

    This class represents a starship engine. */
GEngine::GEngine(int aid, const TEngine& adata)
    : GComponent(aid), data(adata)
{ }

/** Get fuel factor. With the standard fuel consumption formula, this
    is the amount of fuel used by a 100000 kt ship travelling one
    month at the specified speed. The value commonly communicated
    to the user is `getFuelFactor(warp)/(warp*warp)' (fuel per lightyear). */
int
GEngine::getFuelFactor(int warp) const
{
    if (warp == 0)
        return 0;

    ASSERT(warp > 0 && warp < 10);
    return data.fuel[warp-1];
}

/********************************* GHull *********************************/

/** \class GHull
    \brief Starship Hull

    Represents a starship hull. This is information gathered
    together from multiple files: HULLSPEC, HULLFUNC.

    When a ship has a device 'X' which implies 'Y' (see
    GHullFunction::setImpliedDevice), we make sure that every ship
    which receives an 'X' device automatically also has an 'Y' (for
    rule tests). When showing such a ship, and you see a ship which
    has an 'X', the 'Y' should not be displayed. */
GHull::GHull(int aid, const THull& adata)
    : GComponent(aid), data(adata)
{ }

GHull::~GHull()
{ }

TCost
GHull::getCost() const
{
    TCost c = { data.mc, data.t, data.d, data.m };
    return c;
}

/** Set special functions.
    \param hf      Function (hf_xxx)
    \param players Set of players. Only these players will be able to perform
                   the function.
    \param hull    true: this is a hull-specific abilitiy.
                   false: ship-specific (will be attached to ship when built). */
void
GHull::setSpecialFunction(int hf, GPlayerSet players, bool hull)
{
    getSpecialFunctionPtr(hf, hull) = (players.toInteger() & ~1) | hull;
}

/** Add special functions.
    \param hf      Function (hf_xxx)
    \param players Set of players. These players, in addition to those already
                   configured, will be able to perform the function.
    \param hull    true: this is a hull-specific abilitiy.
                   false: ship-specific (will be attached to ship when built). */
void
GHull::addSpecialFunction(int hf, GPlayerSet& players, bool hull)
{
    if (!players.empty())
        getSpecialFunctionPtr(hf, hull) |= (players.toInteger() & ~1) | hull;
}

/** Get definition word for a hull function. Allocates a word if
    necessary. The returned word contains the bitfield which defines
    the players who can access the device: bits 1..11 are players, bit
    0 selects whether it's a hull-specific or ship-specific device (do
    not change it!). */
int16&
GHull::getSpecialFunctionPtr(int hf, bool hull)
{
    for (std::size_t i = 0; i < specials.getSize(); i += 2)
        if (specials[i] == hf && (specials[i+1] & 1) == hull)
            return specials[i+1];

    int16 data[2];
    data[0] = hf;
    data[1] = hull;
    specials.append(data, 2);
    return specials[specials.getSize()-1];
}

/** Compute all implied functions. You should call this after all
    modifications to the hull functions. */
void
GHull::computeImpliedFunctions()
{
    while (1) {
        bool changed = false;
        for (std::size_t i = 0; i < specials.getSize(); i += 2) {
            GHullFunction* special = GHullFunction::getHullFunctionById(specials[i]);
            if (special && special->getImpliedDevice() >= 0) {
                int16 code = specials[i+1];
                int16& ptr = getSpecialFunctionPtr(specials[i], (code & 1));
                if ((ptr & code) != code) {
                    ptr |= code;
                    changed = true;
                }
            }
        }
        if (!changed)
            break;
    }
}

/** Remove all special functions. */
void
GHull::clearSpecialFunctions()
{
    specials.clear();
}

/** Check for special function.
    \param hf     function
    \param player player number
    \param hull   true: query hull-specific ability. Every ship of this
                  type <em>owned</em> by the player has the function.
                  false: query ship-specific ability. Every ship of this
                  type <em>built</em> by the player will have the function. */
bool
GHull::canDoSpecial(int hf, int player, bool hull) const
{
    return getPlayersThatCan(hf, hull).contains(player);
}

/** Get player mask.
    \param hf     function
    \param hull   true: query hull-specific abilities; false: ship-specific
    \returns set all players for whom canDoSpecial() would return true. */
GPlayerSet
GHull::getPlayersThatCan(int hf, bool hull) const
{
    for (std::size_t i = 0; i < specials.getSize(); i += 2)
        if (specials[i] == hf && (specials[i+1] & 1) == hull)
            return GPlayerSet::fromInteger(specials[i+1] & ~1);
    return GPlayerSet();
}

/******************************** Loading ********************************/

/** Load beamspec file. Throws on error. */
static void
loadBeams()
{
    Ptr<Stream> file = spec_file_dir->openFileT("beamspec.dat", Stream::C_READ);

    beam_spec.erase(beam_spec.begin(), beam_spec.end());
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        TBeam data;
        if (!getStructure(*file, data))
            throw FileFormatException(format(_("Error reading `%s'"), "beamspec.dat"));
        beam_spec.push_back(GBeam(i, data));
    }
}

/** Load torpspec file. Throws on error. */
static void
loadTorps()
{
    Ptr<Stream> file = spec_file_dir->openFileT("torpspec.dat", Stream::C_READ);

    torp_spec.erase(torp_spec.begin(), torp_spec.end());
    for (int i = 1; i <= NUM_TORPS; ++i) {
        TTorpedo data;
        if (!getStructure(*file, data))
            throw FileFormatException(format(_("Error reading `%s'"), "torpspec.dat"));
        torp_spec.push_back(GTorpLauncher(i, data));
    }
}

/** Load engspec file. Throws on error. */
static void
loadEngines()
{
    Ptr<Stream> file = spec_file_dir->openFileT("engspec.dat", Stream::C_READ);

    eng_spec.erase(eng_spec.begin(), eng_spec.end());
    for (int i = 1; i <= NUM_ENGINES; ++i) {
        TEngine data;
        if (!getStructure(*file, data))
            throw FileFormatException(format(_("Error reading `%s'"), "engspec.dat"));
        eng_spec.push_back(GEngine(i, data));
    }
}

/** Forget all special functions of all hulls. */
static void
clearSpecials()
{
    for (int i = 1; i < getNumHulls(); ++i)
        getHull(i).clearSpecialFunctions();
}

namespace {
    /** HULLFUNC.TXT parser. */
    class HullfuncParser : public ConfigParser {
        int chosen_hull;
        int chosen_function;
        bool assign_to_hull;
     public:
        HullfuncParser()
            : chosen_hull(-1), chosen_function(-1), assign_to_hull(true)
            { }
        void assignSpecial(int line_id, string_t value, bool byrace);
        void assign(int line_id, string_t name, string_t value, string_t line);
        void error(int line_id, string_t msg);
    };
}

void
HullfuncParser::assign(int line_id, string_t name, string_t value, string_t /*line*/)
{
    if (stringMatch("Initialize", name)) {
        /* "Initialize = Clear" or "Initialize = Default" */
        if (stringMatch("Clear", value)) {
            clearSpecials();
        } else if (stringMatch("Default", value)) {
            clearSpecials();
            GHullFunction::setDefaultSpecials();
        } else {
            error(line_id, _("Invalid argument to `Initialize'"));
        }
    } else if (stringMatch("Hull", name)) {
        /* "Hull = number (Garbage)" or "Hull = Start of name"

           I know this way to parse it is not optimal, but it is
           consistent with PHost */
        int h;
        string_t::size_type pos;
        if ((stringValue(value, h, pos)
             || (pos > 0 && stringValue(value.substr(0, pos), h)))
            && (h > 0 && h <= getNumHulls()))
        {
            chosen_hull = h;
        } else {
            value = strUCase(value);
            chosen_hull = -1;
            for (int i = 1; i <= getNumHulls(); ++i) {
                if (strUCase(getHull(i).getName()).compare(0, value.length(), value) == 0) {
                    chosen_hull = i;
                    break;
                }
            }
            if (chosen_hull < 0)
                error(line_id, _("Invalid hull name"));
        }
    } else if (stringMatch("Function", name)) {
        /* "Function = Number (garbage)" or "Function = Name" */
        int h;
        string_t::size_type pos;
        if ((stringValue(value, h, pos) || (pos > 0 && stringValue(value.substr(0, pos), h)))
            && h >= 0)
        {
            chosen_function = h;
        } else if (GHullFunction* hf = GHullFunction::getHullFunctionByName(value, true)) {
            chosen_function = hf->getId();
        } else {
            error(line_id, _("Invalid hull function"));
        }
    } else if (stringMatch("Racesallowed", name)) {
        assignSpecial(line_id, value, true);
    } else if (stringMatch("Playersallowed", name)) {
        assignSpecial(line_id, value, false);
    } else if (stringMatch("Assignto", name)) {
        assign_to_hull = host.hasShipSpecificFunctions() && stringMatch("Hull", value);
    } else if (stringMatch("Level", name)) {
        // FIXME
    } else {
        error(line_id, _("Invalid keyword"));
    }
}

/** Assign special function.
    \param line_id line number
    \param value   list of players
    \param byrace  if true, the list contains race numbers. Otherwise,
                   player numbers. */
void
HullfuncParser::assignSpecial(int line_id, string_t value, bool byrace)
{
    if (chosen_function < 0) {
        error(line_id, _("No function selected for assignment"));
        return;
    }
    if (chosen_hull < 0) {
        error(line_id, _("No hull selected for assignment"));
        return;
    }

    GPlayerSet players;
    if (host.hasCumulativeHullfunc())
        players = getHull(chosen_hull).getPlayersThatCan(chosen_function, assign_to_hull);
    while (value.length()) {
        bool neg = false;
        if (value[0] == '-') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9') {
                neg = true;
                goto digit;
            }
            players = GPlayerSet();
        } else if (value[0] == '+') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9')
                goto digit;
            players = GPlayerSet::allUpTo(NUM_PLAYERS);
        } else if (value[0] >= '0' && value[0] <= '9') {
         digit:
            string_t::size_type pos;
            int num;
            if (stringValue(value, num, pos)) {
                value.clear();
            } else {
                if (pos && stringValue(value.substr(0, pos), num)) {
                    /* okay */
                    value.erase(0, pos);
                } else {
                    error(line_id, _("Invalid number"));
                    return;
                }
            }
            if (byrace) {
                for (int i = 1; i <= NUM_PLAYERS; ++i) {
                    if (config.PlayerRace[i-1] == num) {
                        if (neg)
                            players -= i;
                        else
                            players += i;
                    }
                }
            } else {
                if (num > 0 && num <= NUM_PLAYERS) {
                    if (neg)
                        players -= num;
                    else
                        players += num;
                } else {
                    error(line_id, _("Invalid player number"));
                    return;
                }
            }
        } else {
            value.erase(0, 1);
        }
    }
    getHull(chosen_hull).setSpecialFunction(chosen_function, players, assign_to_hull);
}

void
HullfuncParser::error(int line_id, string_t msg)
{
    console.write(LOG_WARN, format(_("Line %d: %s"), line_id, msg));
}

/** Load Hull function file. */
static void
loadHullfunc(Ptr<Stream> file, bool in_section)
{
    StandardStream in(file);
    HullfuncParser parser;
    parser.parse(in, "hullfunc", in_section);
}

/** Load hullspec and truehull file. Throws on error.
    \todo Load HULLSPEC.CC(?)
    \pre  configuration has been loaded. */
static void
loadHulls()
{
    Ptr<Stream> file = spec_file_dir->openFileT("hullspec.dat", Stream::C_READ);

    safeClear(hull_spec);
    hull_spec.erase(hull_spec.begin(), hull_spec.end());

    int nhulls = file->getSize() / THull::size;
    if (nhulls < 10)
        throw FileTooShortException("hullspec.dat");
    if (nhulls > MAX_HULLS)
        nhulls = MAX_HULLS;

    for (int i = 1; i <= nhulls; ++i) {
        THull data;
        if (!getStructure(*file, data))
            throw FileFormatException(format(_("Error reading `%s'"), "hullspec.dat"));
        safeAppend(hull_spec, new GHull(i, data));
    }

    file = spec_file_dir->openFileT("truehull.dat", Stream::C_READ);

    if (!getStructure(*file, truehull_spec))
        throw FileFormatException(format(_("Error reading `%s'"), "truehull.dat"));

    for (int p = 0; p < NUM_PLAYERS; ++p)
        for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i)
            if (truehull_spec.hulls[p][i] < 0 || unsigned(truehull_spec.hulls[p][i]) > hull_spec.size())
                throw FileFormatException(format(_("File `%s' is invalid"), "truehull.dat"));

    GHullFunction::readHullfuncDefinitions();
    GHullFunction::setDefaultSpecials();

    if ((file = spec_file_dir->openFileNT("shiplist.txt", Stream::C_READ | Stream::C_TEXT)))
        loadHullfunc(file, false);
    else if ((file = spec_file_dir->openFileNT("hullfunc.txt", Stream::C_READ | Stream::C_TEXT)))
        loadHullfunc(file, true);

    for (int i = 1; i <= getNumHulls(); ++i)
        getHull(i).computeImpliedFunctions();
}

/**************************** Public Interface ***************************/

/** Load Specification Files.
    \pre initGameGlobals() has been called.
    \pre Configuration was loaded. */
void
initSpecification()
{
    loadTorps();
    loadBeams();
    loadEngines();
    loadHulls();
}

/** Get Truehull slot. Returns the position of hull hid in player
    pid's Truehull list, [1, NUM_HULLS_PER_PLAYER]. Returns 0 if he can
    not build that hull. */
int
getTruehullSlot(int pid, int hid)
{
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
        if (getTruehull(pid, i) == hid)
            return i;
    return 0;
}
