/**
  *  \file util/recode.h
  *  \brief Character-set Recoding
  *
  *  Declarations of the interfaces used for character set
  *  translation.
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_RECODE_H
#define PCC_V2_UTIL_RECODE_H

#include <cstddef>
#include "cpluslib/string.h"
#include "defs.h"
#include "cpluslib/types.h"

class Recoder;

/*! \class CharacterSet
    \brief Character Set Handle

    A CharacterSet object is a handle to a character set. You
    generally do not use these objects yourself, you can use
    Recoder::getCharsetByName() to obtain a CharacterSet object for a
    character set given by its name. */
class CharacterSet {
 public:
    class Data;
 private:
    const Data* data;
    friend class Recoder;
 public:
    /** Create character set with given data.
        \param data character set definition. 0 defines an untranslated character set. */
    CharacterSet(const Data* data = 0) : data(data)
        { }

    const char* getName() const;
    const char* getDescription() const;
    uint16 getUnicode(char c) const;
    char getCharacter(uint16 uni) const;

    /** Check whether this is a valid character set.
        \return true iff character set is valid, translation is known.
        false if the character set is unknown and we fall back to untranslated
        mode */
    bool isValid() const
        { return data != 0; }

    /** Compare for equality. */
    bool operator==(CharacterSet rhs) const
        { return data==rhs.data; }

    /** Compare for inequality. */
    bool operator!=(CharacterSet rhs) const
        { return data!=rhs.data; }
};

/*! \class Recoder
    \brief Character Set Recoder

    This class can recode between two character sets. When given two
    character sets, it will try to find a readable translation between
    the two (for example, when it cannot find an accented character in
    the target character set, it attempts to use the unaccented
    character). It will not try to make reversible conversions.

    If either character set is undefined, this will yield an identity
    transformation.

    In addition, this provides two static instances of Recoder for
    translating the "internal" to the "external" character set and
    back. */
class Recoder {
    static const unsigned int TABLE_SIZE = 256;
    char table[TABLE_SIZE];
 public:
    Recoder(CharacterSet from, CharacterSet to);
    Recoder();

    void setTransformation(CharacterSet from, CharacterSet to);
    void setIdentityTransformation();

    string_t recodeString(string_t str) const;
    void     recodeArray(char* data, std::size_t size) const;
    void     recodeCString(char* data) const;

    /** Get recoded value of character.
        \param c character in input character set
        \return character from output character set */
    char recodeCharacter(char c) const
        { return table[(unsigned char) c]; }

    /* Enquiry */
    static CharacterSet  getCharsetByName(const string_t& name);
    static std::size_t   getNumCharsets();
    static CharacterSet  getCharset(std::size_t n);
    
 private:
    /* Static instances */
    static Recoder       intern_to_extern;
    static Recoder       extern_to_intern;
    static CharacterSet  intern_cs;
    static CharacterSet  extern_cs;

 public:
    /** Get recoder to turn internal representation into external. */
    static const Recoder& getI2E()
        { return intern_to_extern; }
    /** Get recoder to turn external representation into internal. */
    static const Recoder& getE2I()
        { return extern_to_intern; }
    /** Get handle to internally-used character set. */
    static CharacterSet getInternCS()
        { return intern_cs; }
    /** Get handle to externally-used character set. */
    static CharacterSet getExternCS()
        { return extern_cs; }
    /** Check whether character is reliable. A character is reliable if
        it survives saving and reloading.
        \param c character in internal character set */
    static bool isReliableChar(char c)
        { return extern_to_intern.recodeCharacter(intern_to_extern.recodeCharacter(c)) == c; }
    
    static void setCharsets(CharacterSet intrn, const CharacterSet extrn);
};

#endif
